from sklearn.metrics import confusion_matrix
import numpy as np


def calculate_se_sp(y_true, y_pred):
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
    specificity = tn / (tn + fp)
    sensitivity = tp / (tp + fn)

    return sensitivity, specificity


def specificity_fixed_sensitivity(y_true, y_pred, min_se=0.95):
    initial_step = (y_pred.max() - y_pred.min()) / 10
    min_val = y_pred.min()

    T = min_val + initial_step

    y_pred_class = np.zeros(len(y_pred))
    y_pred_class[y_pred > T] = 1

    se, _ = calculate_se_sp(y_true, y_pred_class)

    while se > min_se:
        T += initial_step

        y_pred_class = np.zeros(len(y_pred))
        y_pred_class[y_pred > T] = 1

        se, _ = calculate_se_sp(y_true, y_pred_class)

    while se <= min_se:
        T -= 0.01
        y_pred_class[y_pred > T] = 1

        se, sp = calculate_se_sp(y_true, y_pred_class)

    return {
        "se": se,
        "sp": sp,
        "g_mean": np.sqrt(se * sp),
        "T": T,
    }


def calculate_se_sp_gmeans_with_threshold(y_true, y_pred, T=0):
    y_pred_class = np.zeros(len(y_pred))
    y_pred_class[y_pred > T] = 1
    se, sp = calculate_se_sp(y_true, y_pred_class)

    return {
        "se": se,
        "sp": sp,
        "g_mean": np.sqrt(se * sp),
    }


def calculate_gmean(y_true, y_pred):
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
    specificity = tn / (tn + fp)
    sensitivity = tp / (tp + fn)
    return np.sqrt(specificity * sensitivity)



import numpy as np
from sklearn.svm import SVC
from .conformal_transformation import conformal_transform_kernel, calculate_tau_squared
from .distance import hyperspace_l2_distance_squared
from .kernel import laplacian_kernel


class AFSCTSvm:

    def __init__(
        self, C=1, class_weight="balanced", neg_eta=None, pos_eta=None, kernel=None, ignore_outlier_svs=True, probability=False, random_state=0
    ):
        self.C = C
        self.class_weight = class_weight
        self.kernel = kernel
        self.probability = probability
        self.random_state = random_state
        self.neg_eta = neg_eta
        self.pos_eta = pos_eta

        if self.kernel is None:
            self.kernel = laplacian_kernel

        self.ignore_outlier_svs = ignore_outlier_svs

    def fit(self, X, y, sample_weight=None):
        computed_conformal_transform_kernel = self.fit_conformal_transformed_kernel(X, y, sample_weight)

        if sample_weight is not None:
            self.svm = SVC(
                C=self.C,
                kernel="precomputed", # set class weight to None when using sample weights
                probability=self.probability,
                random_state=self.random_state
            )
            self.svm.fit(computed_conformal_transform_kernel, self.y_train, sample_weight=sample_weight)
        else:
        
            self.svm = SVC(
                C=self.C,
                class_weight=self.class_weight,
                kernel="precomputed",
                probability=self.probability,
                random_state=self.random_state

            )
            self.svm.fit(computed_conformal_transform_kernel, self.y_train)

    def fit_conformal_transformed_kernel(self, X, y, sample_weight=None):
        self.X_train = X
        self.y_train = y

        if sample_weight is not None:
            self.svm = SVC(
                C=self.C,
                kernel="precomputed",
                probability=self.probability,
                random_state=self.random_state

            )

            computed_kernel = self.kernel(self.X_train, self.X_train)
            self.svm.fit(computed_kernel, self.y_train, sample_weight=sample_weight)

        else:


            self.svm = SVC(
                C=self.C,
                class_weight=self.class_weight,
                kernel="precomputed",
                probability=self.probability,
                random_state=self.random_state

            )

            computed_kernel = self.kernel(self.X_train, self.X_train)
            self.svm.fit(computed_kernel, self.y_train)

        support_vectors_pos, support_vectors_neg = (
            self.extract_separate_support_vectors()
        )
        tau_squareds = self.calculate_tau_squared()
        support_vectors = np.vstack((support_vectors_pos, support_vectors_neg))

        self.tau_squareds = tau_squareds
        self.support_vectors = support_vectors

        computed_conformal_transform_kernel = conformal_transform_kernel(
            self.X_train, self.X_train, computed_kernel, support_vectors, tau_squareds
        )

        return computed_conformal_transform_kernel
    
    def get_computed_conformal_transform_kernel(self, X):
        computed_kernel = self.kernel(X, self.X_train)
        computed_conformal_transform_kernel = conformal_transform_kernel(
            X, self.X_train, computed_kernel, self.support_vectors, self.tau_squareds
        )
        return computed_conformal_transform_kernel

    def predict(self, X):
        computed_conformal_transform_kernel = self.get_computed_conformal_transform_kernel(X)
        return self.svm.predict(computed_conformal_transform_kernel)

    def predict_proba(self, X):
        computed_conformal_transform_kernel = self.get_computed_conformal_transform_kernel(X)
        return self.svm.predict_proba(computed_conformal_transform_kernel)
    
    def decision_function(self, X):
        computed_kernel = self.kernel(X, self.X_train)
        computed_conformal_transform_kernel = conformal_transform_kernel(
            X, self.X_train, computed_kernel, self.support_vectors, self.tau_squareds
        )
        return self.svm.decision_function(computed_conformal_transform_kernel)

    def extract_separate_support_vectors(self):
        support_vectors = self.X_train[self.svm.support_]
        support_vectors_class = self.y_train[self.svm.support_]
        support_vectors_pos = support_vectors[np.where(support_vectors_class == 1)]
        support_vectors_neg = support_vectors[np.where(support_vectors_class == 0)]

        if self.ignore_outlier_svs:
            
            support_vectors_pos_pred = self.svm.predict(
                self.kernel(support_vectors_pos, self.X_train)
            )
            support_vectors_pos = support_vectors_pos[support_vectors_pos_pred == 1]

            support_vectors_neg_pred = self.svm.predict(
                self.kernel(support_vectors_neg, self.X_train)
            )
            support_vectors_neg = support_vectors_neg[support_vectors_neg_pred == 0]

        self.support_vectors_pos = support_vectors_pos
        self.support_vectors_neg = support_vectors_neg
        return support_vectors_pos, support_vectors_neg

    def calculate_tau_squared(self):
        distances = hyperspace_l2_distance_squared(
            self.support_vectors_pos, self.support_vectors_neg, self.kernel
        )
        tau_squared_pos = calculate_tau_squared(distances, self.pos_eta)
        tau_squared_neg = calculate_tau_squared(distances.T, self.neg_eta)
        tau_squareds = np.concatenate((tau_squared_pos, tau_squared_neg))
        return tau_squareds



import numpy as np


def hyperspace_l2_distance_squared(X, Y, kernel):
    K_X = kernel(X, X)
    K_Y = kernel(Y, Y)
    K_XY = kernel(X, Y)
    K_X = np.diag(K_X)
    K_Y = np.diag(K_Y)
    K_X = np.tile(K_X, (K_XY.shape[1], 1)).T
    K_Y = np.tile(K_Y, (K_XY.shape[0], 1))
    return K_X + K_Y - 2 * K_XY



from scipy.spatial.distance import cdist
import numpy as np
from sklearn.gaussian_process.kernels import Kernel


def laplacian_kernel(XA, XB, gramma=1.0):
    l1_norm = cdist(XA, XB, "minkowski", p=1)
    return np.exp(-gramma * l1_norm)


class RBFKernel(Kernel):

    def __init__(self, gamma=1):
        self.gamma = gamma

    def __call__(self, X, Y=None):
        if Y is None:
            Y = X
        pairwise_sq_dists = cdist(X, Y, "sqeuclidean")
        K = np.exp(-self.gamma * pairwise_sq_dists)
        return K


class LaplacianKernel(Kernel):

    def __init__(self, gamma=1):
        self.gamma = gamma

    def __call__(self, X, Y=None):
        if Y is None:
            Y = X
        l1_norm = cdist(X, Y, "minkowski", p=1)
        return np.exp(-self.gamma * l1_norm)



import numpy as np
from scipy.spatial.distance import cdist


def conformal_transform_kernel(X, Y, computed_kernel, support_vectors, tau_squareds):
    D_X = calculate_D(X, support_vectors, tau_squareds)
    D_Y = calculate_D(Y, support_vectors, tau_squareds)
    D_X = D_X.reshape((-1, 1))
    D_Y = D_Y.reshape((1, -1))
    D_XY = np.matmul(D_X, D_Y)
    return np.multiply(D_XY, computed_kernel)


def calculate_D(X, support_vectors, tau_squared):
    l1_dist = cdist(X, support_vectors, "minkowski", p=1)
    return np.exp(-l1_dist / tau_squared).sum(axis=1)


def calculate_tau_squared(distance_mat, eta=None):
    M = (np.max(distance_mat, axis=1) + np.min(distance_mat, axis=1)) / 2
    mask = distance_mat < M.reshape(-1, 1)
    masked_distance = distance_mat * mask

    if eta is None:

        if distance_mat.shape[0] < distance_mat.shape[1]:
            eta = 1
        else:
            eta = distance_mat.shape[1] / distance_mat.shape[0]
    
    tau_squared = masked_distance.sum(axis=1) / mask.sum(axis=1)

    return tau_squared * eta
